@@ -186,7 +186,8 @@ class TableBase(metaclass=ABCMeta):
         key: t.Optional[t.Dict[str, t.Any]] = None,
         filters: t.Optional[t.Dict[str, t.Any]] = None,
         limit: t.Optional[int] = None,
-        next: t.Optional[str] = None
+        next: t.Optional[str] = None,
+        index: t.Optional[str] = None
     ) -> t.Dict[str, t.Any]:
         """Perform key based query with optional exact match filters
 
@@ -196,6 +197,7 @@ class TableBase(metaclass=ABCMeta):
             filters: optional dictionary of key=value to query and filter on
             limit: query limit
             next: pagination token
+            index: name of index to use (dynamodb only)
 
         Returns:
             dictionary containing 'items' and 'next' pagination token
@@ -489,7 +491,7 @@ def add_audit(item: t.Dict, update: bool, user: str) -> t.Dict:
         item
 
     """
-    camel_case = os.environ.get('ABNOSQL_AUDIT_CAMELCASE', 'TRUE') == 'TRUE'
+    camel_case = os.environ.get('ABNOSQL_CAMELCASE', 'TRUE') == 'TRUE'
     by_attr = 'By' if camel_case else '_by'
     date_attr = 'Date' if camel_case else '_date'
     dt_iso = datetime.now(timezone.utc).strftime('%Y-%m-%dT%H:%M:%SZ')
@@ -505,6 +507,39 @@ def add_audit(item: t.Dict, update: bool, user: str) -> t.Dict:
     return item
 
 
+def add_change_meta(item: t.Dict, event_source: str, event_name: str) -> t.Dict:
+    """Add changeMetadata object to item containing eventName and eventSource
+
+    Args:
+
+        item: item dict
+        event_source: str
+        event_name: str - INSERT, MODIFY or REMOVE
+
+    Returns:
+        item
+
+    """
+    event_name = event_name.upper()
+    if event_name not in ['INSERT', 'MODIFY', 'REMOVE']:
+        return item
+    camel_case = os.environ.get('ABNOSQL_CAMELCASE', 'TRUE') == 'TRUE'
+    meta_attr = 'changeMetadata' if camel_case else 'change_metadata'
+    source_attr = 'eventSource' if camel_case else 'event_source'
+    name_attr = 'eventName' if camel_case else 'event_name'
+    # cosmos delete_item() adds REMOVE so don't add if already present in item
+    existing_event = item.get(meta_attr, {}).get(name_attr)
+    if existing_event == 'REMOVE':
+        return item
+    item.update({
+        meta_attr: {
+            source_attr: event_source,
+            name_attr: event_name
+        }
+    })
+    return item
+
+
 def parse_connstr():
     connstr = os.environ.get('ABNOSQL_DB')
     if connstr:
@@ -546,10 +581,17 @@ def table(
                 database = _database
                 break
 
+    if database is None:
+        raise ex.PluginException('table plugin database not defined')
+
     pm = plugin.get_pm('table')
     module = pm.get_plugin(database)
     if module is None:
         raise ex.PluginException(f'table.{database} plugin not found')
+    if hasattr(module, 'MISSING_DEPS'):
+        raise ex.PluginException(
+            f'table.{database} plugin missing dependencies'
+        )
     if not isinstance(config, dict):
         config = {}
     _module = module.Table(pm, name, config)
@@ -577,7 +619,7 @@ def table(
         # check required attrs
         missing = [
             _ for _ in ['attrs', 'key_attrs']
-            if not isinstance(kcfg[_], list) or len(kcfg[_]) == 0
+            if not isinstance(kcfg.get(_), list) or len(kcfg[_]) == 0
         ]
         if len(missing):
             raise ex.ConfigException(
